
MIN_N_WARM = 101   # Library will fail if history is too small

DIVINE_MODEL = dict(confidence_interval = 34.134*2,   # So it is +/- one standard deviation
                    seasonal_periods = [7, 14., 28., 30., 90., 120., 182., 365.],
                     optimise_trend_season_features=True,
                    trend_order=[0, 1]
                    )



try:
    import pyflux as pf
except ImportError:
    class Mock:
        ARIMA = None
    pf = Mock()
    print('pyflux is off the list while the package on PyPI is broken. You can install from git')
    print('pip install git+https://github.com/RJT1990/pyflux.git')

import pandas as pd
from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous
from timemachines.skatertools.utilities.conventions import to_int_log_space, initialize_buffers, split_exogenous, update_buffers, \
    R_TYPE

# TODO: Span models at https://pyflux.readthedocs.io/en/latest/gpnar.html with r


def flux_hyperparams(s, r: R_TYPE):
    # Interpret hyper-parameters - r is only used once at initiation of model
    integ_bounds, ar_bounds, ma_bounds =(0,2), (0.55, 2.45), (0.66, 10.33)
    s['n_burn'] = 25
    s['alpha'] = 0.25  # Defines confidence interval
    s['buffer_len'] = 5000
    s['integ'], s['ar'], s['ma'] = to_int_log_space(r, bounds=[integ_bounds, ar_bounds, ma_bounds])
    s['family'] = pf.Normal()
    return s


def flux_or_last_value(s, k:int, exog:[float], y0:float):
    # We fit each time
    if s.get('model') is None:
        return y0, s, None
    else:
        y_hat = s['model'].predict(h=k)
        x = y_hat.values[0][0]
        w = None
        return x, s, w


def flux_auto(y, s, k, a, t, e, r):
    """ One way to use flux package

            - Contemporaneous y[1:] variables are used as exogenous 'X' in pmdarima
            - This only works for k=1

        :returns: x, s', w
    """
    if s is None:
        s = dict()
        s = flux_hyperparams(s=s,r=r)
        s = initialize_buffers(s=s,y=y)

    if y is not None:
        # Process observation and return prediction
        assert isinstance(y, float) or len(y) == s['dim'], ' Cannot change dimension of input in flight '
        y0, exog = split_exogenous(y=y, dim=s['dim'])
        s = update_buffers(s=s, a=a, exog=exog, y0=y0)
        if True:  # Always fit prior to prediction
            none_, s, _ = flux_auto(y=None, s=s, k=k, a=a, t=t, e=e, r=r)  # Fit the model
            assert none_ is None
        return flux_or_last_value(s=s,k=k,exog=exog,y0=y0)

    if y is None:
        if len(s.get('buffer'))<s['n_burn']:
            s['model'] = None
        else:
            data = pd.DataFrame(columns=['y'], data=s.get('buffer'))
            s['model'] = pf.ARIMA(data=data, ar=s['ar'], ma=s['ma'], target='y', family=s['family'])
            _ = s['model'].fit("MLE")
        return None, s, None  # Acknowledge that a fit was requested by returning x=None, w=None


if __name__ == '__main__':
    err = prior_plot_exogenous(f=flux_auto, k=1, n=200, r=0.05)
    pass



from timemachines.skatertools.utilities.conventions import Y_TYPE, from_space, A_TYPE
from timemachines.skatertools.evaluation.evaluators import evaluate_mean_squared_error
import traceback
import time


# Find the optimal hyper-parameter r for a skater

# TODO:Tidy this up

try:
    from humpday.optimizers.alloptimizers import OPTIMIZERS
    using_humpday = True
except:
    using_humpday = False



def infer_r_dim(skater_name):
    """ A convention for skater names """
    return 3 if '_r3' in skater_name else 2 if '_r2' in skater_name else 1 if '_r1'in skater_name else None


def optimal_r(f, y:[Y_TYPE],
              k:int, a:[A_TYPE], t=None, e=None,
              evaluator=None, optimizer=None,
              n_trials=None, n_dim=None,  # Which dimension to search in
              n_burn:int=None,
              test_objective_first=True)->(float,float,dict):
    """
          Returns:  best_r: float
                    best_val: float
                    info: {best_rn: interpretation in n_dim space of r,
                           feval_count: number of actual objective evaluations,
                          }

    """
    if not using_humpday:
        raise Exception('You must manually install humpday, with pip install humpday, in order to use this')
    if evaluator is None:
        evaluator = evaluate_mean_squared_error

    if n_dim is None:
        n_dim = infer_r_dim(f.__name__)
        assert n_dim is not None, "Hyper-parameter dimension was not supplied and could not be inferred from "+str(f.__name__)

    assert n_burn is not None, "Must supply n_burn"

    def objective(u:[float]):
        r = from_space(u)
        return evaluator(f=f, y=y, k=k, a=a, t=t, e=r, r=r, n_burn=n_burn)

    if test_objective_first:
        start_time = time.time()
        a_test = objective(u=[0.5]*n_dim)  # Fail fast with easier trace
        elapsed = time.time()-start_time
        print('One evaluation took '+ str(round(elapsed/60,1))+' minutes.')
        print('Come back in '+str(n_trials*elapsed/60)+' minutes.')
    best_val, best_rn, feval_count = optimizer(objective, n_trials=n_trials, n_dim=n_dim, with_count=True)
    best_r = from_space(best_rn)
    info = {'best_rn':best_rn,'feval_count':feval_count}
    return best_r, best_val, info






def optimize_proph_against_hospital(k=10, n_extra=100):
    """ Illustrative example

           n_extra is roughly the number of times the prophet model will be called

    """
    assert using_humpday,'pip install humpday'
    from timemachines.skatertools.evaluation.evaluators import evaluate_mean_squared_error
    from timemachines.skatertools.data.real import hospital_with_exog
    from timemachines.skaters.proph.prophparams import PROPHET_META
    from humpday.optimizers.optunacube import optuna_tpe_cube
    evaluator = evaluate_mean_squared_error
    n_burn = PROPHET_META['n_warm'] + k + 1
    y, a = hospital_with_exog(k=k, n=n_burn+n_extra, offset=True)

    from timemachines.skaters.proph.prophskaterssingular import fbprophet_univariate, fbprophet_known, fbprophet_recursive
    from timemachines.skaters.simple.movingaverage import EMA_SKATERS
    skaters = [fbprophet_univariate, fbprophet_known, fbprophet_recursive, fbprophet_recursive] + EMA_SKATERS

    for f in skaters:
        mn_err = evaluator(f=f, y=y, k=k, a=a, n_burn=n_burn)
        print(f.__name__ + ' : '+ str(mn_err))

    # Finally, try to optimize prophet with hyper-parameters
    from timemachines.skaters.proph.prophskaterssingular import fbprophet_exogenous_r2
    f = fbprophet_exogenous_r2
    n_dim = 2
    best_r, best_val, info = optimal_r(f=f, y=y, k=k, a=a, evaluator=evaluator,
                                 optimizer=optuna_tpe_cube, n_burn=n_burn,
                                 n_trials=50, n_dim=n_dim,
                                 test_objective_first=True)
    from pprint import pprint
    pprint(info)
    print('Best r is ' + str(best_r))
    print('Best value is '+str(best_val))
    from timemachines.skaters.proph.prophparams import prophet_param
    from timemachines.skatertools.utilities.conventions import to_space
    rn = to_space(best_r)
    pprint({'changepoint prior':prophet_param(param_name='changepoint_prior_scale',u=rn[0]),
           'seasonality prior':prophet_param(param_name='seasonality_prior_scale',u=rn[0])})

    # But out of sample...
    # TODO


def select_working_combinations():
    """ Tedious examination of which optimizers work against different evaluators """
    from timemachines.skatertools.data.synthetic import brownian_with_exogenous
    from timemachines.skatertools.evaluation.evaluators import EVALUATORS
    broken = list()
    k = 1
    from timemachines.skaters.simple.movingaverage import empirical_ema_r1
    f = empirical_ema_r1
    for optimizer in OPTIMIZERS:
        print(' ')
        print(optimizer.__name__)
        for evaluator in EVALUATORS:
            try:
                print(evaluator.__name__, optimal_r(f=f, y=brownian_with_exogenous(n=40), k=k, a=None,
                                                    n_trials=5, n_dim=3, n_burn=20, optimizer=optimizer,
                                                    evaluator=evaluator, with_count=True))
            except Exception as e:
                traceback.print_tb()
                broken.append((optimizer.__name__, evaluator.__name__))
    print(' ')
    print('Broken : ')
    print(broken)


if __name__=='__main__':
    optimize_proph_against_hospital()




from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE
from typing import Any
from timemachines.skatertools.composition.residualcomposition import residual_chaser_factory
from timemachines.skaters.simple.hypocratic import quickly_hypocratic, slowly_hypocratic


def quickly_moving_hypocratic_residual_factory(y :Y_TYPE, s:dict, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None,
                                               f=None, r=None)->([float] , Any , Any):
    """ A simple way to convert a skater into one whose residuals are cautiously chased with a quickly moving average

           f : skater
           r : skater's hyper-params, if any

    """
    return residual_chaser_factory(y=y,s=s,k=k,a=a,t=t,e=e,f1=f,r1=r,f2=quickly_hypocratic,r2=r)


def slowly_moving_hypocratic_residual_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None,
                                              e: E_TYPE = None,
                                              f=None, r=None) -> ([float], Any, Any):
    """ A simple way to convert a skater into one whose residuals are cautiously chased with a slowly moving average

           f : skater
           r : skater's hyper-params, if any

    """
    return residual_chaser_factory(y=y, s=s, k=k, a=a, t=t, e=e, f1=f, r1=r, f2=slowly_hypocratic, r2=r)


from timemachines.skaters.allskaters import SKATERS, skater_from_name
from timemachines.skating import prior, posterior, residuals
from timemachines.skatertools.visualization.priorplot import prior_plot, prior_plot_exogenous


from timemachines.skaters.divine.divineskaterfactory import divinity_univariate_factory
from timemachines.skatertools.composition.residualshypocratic import quickly_moving_hypocratic_residual_factory, slowly_moving_hypocratic_residual_factory


def divine_univariate(y,s,k,a=None,t=None,e=None):
    return divinity_univariate_factory(y=y, s=s, k=k, a=a, t=t, e=e)


def divine_univariate_hypocratic_slow(y,s,k,a=None,t=None,e=None):
    return slowly_moving_hypocratic_residual_factory(f=divine_univariate, y=y,s=s,k=k,a=a,t=t,e=e)


def divine_univariate_hypocratic_fast(y,s,k,a=None,t=None,e=None):
    return quickly_moving_hypocratic_residual_factory(f=divine_univariate, y=y,s=s,k=k,a=a,t=t,e=e)


DIVINE_SKATERS = [ divine_univariate, divine_univariate_hypocratic_slow, divine_univariate_hypocratic_fast ]


from timemachines.skaters.proph.prophiskaterfactory import using_prophet
if using_prophet:
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE
    from timemachines.skatertools.composition.residualshypocratic import quickly_moving_hypocratic_residual_factory
    from timemachines.skatertools.composition.residualcomposition import residual_chaser_factory
    from timemachines.skaters.proph.prophskaterssingular import fbprophet_univariate, fbprophet_exogenous, fbprophet_cautious


    def fbprophet_exogenous_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Chase residuals, somewhat cautiously using, quickly moving average """
        return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=None, f=fbprophet_exogenous)


    def fbprophet_univariate_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Chase residuals, somewhat cautiously using, quickly moving average """
        return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=None, f=fbprophet_univariate)


    def fbprophet_cautious_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Chase residuals, somewhat cautiously using, quickly moving average """
        return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=None, f=fbprophet_cautious)


    def fbprophet_exogenous_exogenous(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        return residual_chaser_factory(y=y, s=s, k=k, a=a, t=t, e=e, f1=fbprophet_exogenous, f2=fbprophet_exogenous  )


    def fbprophet_univariate_univariate(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        return residual_chaser_factory(y=y, s=s, k=k, a=a, t=t, e=e, f1=fbprophet_univariate, f2=fbprophet_univariate )


    def fbprophet_univariate_univariate_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        return residual_chaser_factory(y=y, s=s, k=k, a=a, t=t, e=e, f1=fbprophet_univariate, f2=fbprophet_univariate_hypocratic )


    PROPHET_SKATERS_COMPOSED = [fbprophet_exogenous_hypocratic, fbprophet_univariate_hypocratic, fbprophet_cautious_hypocratic,
                                fbprophet_exogenous_exogenous, fbprophet_univariate_univariate_hypocratic]
else:
    PROPHET_SKATERS_COMPOSED = []



if __name__ == '__main__':
    from timemachines.skatertools.data.real import hospital_with_exog
    from timemachines.skatertools.visualization.priorplot import prior_plot
    import matplotlib.pyplot as plt
    k = 1
    y, a = hospital_with_exog(k=k, n=450, offset=True)
    f = fbprophet_univariate_hypocratic
    err2 = prior_plot(f=f, k=k, y=y, n=450, n_plot=50)
    print(err2)
    plt.show()
    pass




NPROPHET_META={'n_warm':400,  # Default minimum number of data points before fitting
              'freq':'H',    # Sampling frequency used if none is supplied, or t not supplied
              'n_max':500}   # Maximum length of training data


NPROPHET_MODEL = dict(n_lags=20,
    changepoints_range=0.95,
    n_changepoints=30,
    weekly_seasonality=False,
    batch_size=64,
    epochs=10,
    learning_rate=1.0)


from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any, List
from timemachines.skatertools.ensembling.precisionweightedskater import precision_weighted_skater


def ensemble_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None,
                     fs: List = None, rs: List = None, g=None, r=None, include_std=True) -> ([float], Any, Any):
    """ Ensembles *only* the k-step ahead

          fs  - list of skaters
          rs  - list of hyper-params, if any
          g   - exogenous skater
          r   - hyper-param for g, if any
          include_std - bool. If True, will add x_std to the exogenous variables sent to g

    """
    if not s.get('s_fs'):
        s = {'s_fs': [{} for _ in fs],
             's_g': {},
             'n_obs':0}

    if y is None:
        return None, None, s
    else:
        # Apply models, keeping only the point estimate
        xjs = list()
        rs = rs or [None for _ in fs]
        for j, (f, r) in enumerate(zip(fs, rs)):
            if r is not None:
                xj, xj_std, s['s_fs'][j] = f(y=y, s=s['s_fs'][j], k=k, a=a, t=t, e=e, r=r)
            else:
                xj, xj_std, s['s_fs'][j] = f(y=y, s=s['s_fs'][j], k=k, a=a, t=t, e=e)
            xjs.append(xj[-1])
            if include_std:
                xjs.append(xj_std[-1])

        s['n_obs']+=1
        if s['n_obs']<10:
            return [wrap(y)[0]]*k, [wrap(y)[0]]*k, s
        else:
            y_extend = [wrap(y)[0]] + xjs
            if r is None:
                x, x_std, s['s_g'] = g(y=y_extend, s=s['s_g'], k=k, a=a, t=t, e=e)
            else:
                x, x_std, s['s_g'] = g(y=y_extend, s=s['s_g'], k=k, a=a, t=t, e=e, r=r)
            return x, x_std, s


def precision_weighted_ensemble_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None,
                     fs: List = None, rs: List = None, r=None) -> ([float], Any, Any):
    """
             r -   determines the exponent
                   r=0.5 corresponds to simple weighting, and is the default
    """
    g = precision_weighted_skater
    return ensemble_factory(y=y,s=s,k=k,a=a,t=t,e=e,fs=fs,rs=rs,g=g,r=r)




from statistics import mode, StatisticsError
import numpy as np


def approx_dt(t:[float], default_dt=60 ):
    """ Crude estimate of the typical time between arrivals
          t: list of epoch times
    """
    if len(t) > 5:
        return approx_mode([abs(dt) for dt in np.diff(list(t))]) or default_dt
    else:
        return default_dt


def approx_mode(xs, ndigits=0):
    """ Mode of rounded numbers, or None """
    xr = [round(x, ndigits) for x in xs]
    try:
        return mode(xr)
    except StatisticsError:
        return None


from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE
from timemachines.skaters.simple.movingaverage import empirical_ema_r1
from timemachines.skatertools.utilities.conventions import wrap

# Skaters that shrink towards zero


def hypocratic_ema_r1(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None, r:R_TYPE=None):
    """
         r :  moving average parameter  (e.g. 0.75 is fast, 0.95 is slow)
    """
    y0 = wrap(y)[0]
    assert r is not None
    x, x_std, s = empirical_ema_r1(y=y0,s=s,k=k,a=a,t=t,e=e,r=r)

    def hypocratic(x:float, x_std:float, confidence=0.5):
        """ Shrink residual prediction towards zero """
        import math
        if abs(x_std)<1e-6 or abs(x)<1e-3*x_std:
            return 0.0
        else:
            return x*math.tanh(confidence*abs(x)/(3*x_std))

    x_resid = [ hypocratic(xi,x_std) for xi,x_std in zip(x,x_std) ]
    return x_resid, x_std, s


def slowly_hypocratic(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None):
    return hypocratic_ema_r1(y=y,s=s,k=k,a=a,t=t,e=e,r=0.95)


def quickly_hypocratic(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None):
    return hypocratic_ema_r1(y=y,s=s,k=k,a=a,t=t,e=e,r=0.75)



if __name__=='__main__':
    from timemachines.skatertools.data.real import hospital_with_exog
    from timemachines.skatertools.evaluation.evaluators import evaluate_mean_absolute_error

    k = 3
    y, a = hospital_with_exog(k=k,n=500)
    f = slowly_hypocratic
    err1 = evaluate_mean_absolute_error(f=f, k=k, y=y, a=a, n_burn=50)



import pandas as pd
from typing import List, Tuple, Any
from timemachines.skatertools.utilities.conventions import wrap
from timemachines.skaters.nproph.nprophparams import NPROPHET_MODEL, NPROPHET_META

try:
    from neuralprophet import NeuralProphet
    using_neuralprophet = True
except ImportError:
    using_neuralprophet = False

if using_neuralprophet:

    def nprophet_iskater_factory(y: [[float]], k: int, a: List = None, t: List = None, e=None, freq: str = None, n_max=1000,
                                recursive: bool = False, model_params: dict = None, return_forecast=True):
        # For now we keep it simple. Will add to this over time
        y0s = [wrap(yi)[0] for yi in y]
        x, x_std, forecast,m = nprophet_fit_and_predict_simple(y=y0s,k=k,freq=freq,model_params=model_params)
        return (x, x_std, forecast, m) if return_forecast else (x, x_std)


    def nprophet_fit_and_predict_simple(y: [float], k: int, freq: str = None, model_params: dict = None) -> Tuple[
        List, List, Any, Any]:
        """ Simpler wrapper for testing - univariate only """
        assert isinstance(y[0],float)
        freq = freq or NPROPHET_META['freq']
        used_params = NPROPHET_MODEL
        if model_params:
            used_params.update(model_params)

        if len(y)<used_params['n_lags']:
            x = [wrap(y)[0]]*k
            x_std = [1.0]*k
            return x, x_std, None, None
        else:
            model = NeuralProphet(**used_params)
            df = pd.DataFrame(columns=['y'], data=y)
            df['ds'] = pd.date_range(start='2021-01-01', periods=len(y), freq=freq)
            metrics = model.fit(df, freq='5min')
            future = model.make_future_dataframe(df, periods=k)
            forecast = model.predict(future)
            x = forecast['yhat1'].values[-k:]
            x_std = [1.0]*k
            return x, x_std, forecast, model


if __name__=='__main__':
    from timemachines.skatertools.data.real import hospital

    k = 3
    n = 500
    y = hospital(n=n)[-200:]
    x, x_std, forecast, m = nprophet_iskater_factory(y=y, k=k)
    assert len(x) == k
    x1, x_std1, forecast1, m1 = nprophet_fit_and_predict_simple(y=y, k=k)
    if True:
        m.plot(forecast)
        m1.plot(forecast1)
        import matplotlib.pyplot as plt
        plt.show()


try:
    from microprediction import MicroReader
    have_micro = True
except ImportError:
    have_micro = False

import random

# Live data streams at www.microprediction.com

if have_micro:
    def random_regular(min_len=500):
        """ Randomly selected univariate series
        :return:  y, t, url
        """
        mr = MicroReader()
        names = mr.get_stream_names()
        okay = False
        while not okay:
            name = random.choice(names)
            n_obs = len(mr.get_lagged_values(name=name,count=10000))
            okay = n_obs > min_len and '~' not in name
        url = 'https://www.microprediction.org/stream_dashboard.html?stream=' +name.replace('.json','')
        return name, url


    def stream_data(name:str,n_obs:int):
        """ values and times for a univariate stream """
        mr = MicroReader()
        lagged_values, lagged_times = mr.get_lagged_values_and_times(name=name, count=n_obs)
        y, t = list(reversed(lagged_values)), list(reversed(lagged_times))
        return y, t


    def random_regular_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
        :returns:  y, t
        """
        name, url = random_regular(min_len=n_obs)
        return stream_data(name=name,n_obs=n_obs)


if __name__=='__main__':
    if have_micro:
        name, url = random_regular(min_len=900)
        print(url)


from pydlm import autoReg, dlm


class fixedAutoReg(autoReg):
    # A hack to make DLM auto work, the version in PyPI at time of writing throws error trying to append

    def __init__(self,*arg,**kwargs):
        super().__init__(*arg,**kwargs)

    def appendNewData(self, data):
        """ AutoReg append new data automatically with the main time series. Nothing
        needs to be done here.
        """
        return


class quietDlm(dlm):
    # Same as DLM but without priting

    def __init__(self,*arg,**kwargs):
        super().__init__(*arg,**kwargs)

    # Append new data or features to the dlm
    def append(self, data, component='main'):
        """ Append the new data to the main data or the components (new feature data)

        Args:
            data: the new data
            component: the name of which the new data to be added to.\n
                       'main': the main time series data\n
                       other omponent name: add new feature data to other
                       component.

        """
        # initialize the model to ease the modification
        if not self.initialized:
            self._initialize()

        # if we are adding new data to the time series
        if component == 'main':
            # add the data to the self.data
            self.data.extend(list(data))

            # update the length
            self.n += len(data)
            self.result._appendResult(len(data))

            # update the automatic components as well
            for component in self.builder.automaticComponents:
                comp = self.builder.automaticComponents[component]
                comp.appendNewData(data)

        # if we are adding new data to the features of dynamic components
        elif component in self.builder.dynamicComponents:
            comp = self.builder.dynamicComponents[component]
            comp.appendNewData(data)

        else:
            raise NameError('Such dynamic component does not exist.')



from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any
from timemachines.skatertools.components.parade import parade

from timemachines.skaters.nproph.nprophetiskaterfactory import using_neuralprophet
if using_neuralprophet:
    from timemachines.skaters.nproph.nprophetiskaterfactory import nprophet_iskater_factory
    from timemachines.skatertools.utilities.nonemath import nonecenter
    from timemachines.skaters.proph.prophparams import PROPHET_META, prophet_params
    from timemachines.skatertools.utilities.nonemath import nonecast
    import sys
    import logging
    import numpy as np

    logging.disable(sys.maxsize)
    logging.getLogger('fbprophet').setLevel(logging.ERROR)


    def nprophet_skater_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None,
                                 emp_mass: float = 0.0, emp_std_mass: float = 1.0,
                                 freq=None, recursive: bool = False,
                                 model_params: dict = None,    # n_lags ...
                                 n_max: int = None) -> ([float], Any, Any):
        """ Neural prophet skater with running prediction error moments
            Hyper-parameters are explicit here, whereas they are determined from r in actual skaters.
            Params of note:

                 a: value of known-in-advance vars k step in advance (not contemporaneous with y)
                 emp_mass: How much to auto-correct bias
                 emp_std_mass: How much to use empirical std versus model

        """

        assert 0 <= emp_mass <= 1
        assert 0 <= emp_std_mass <= 1

        if freq is None:
            freq = PROPHET_META['freq']
        if n_max is None:
            n_max = PROPHET_META['n_max']

        y = wrap(y)
        a = wrap(a)

        if not s.get('y'):
            s = {'p': {},     # parade
                 'y': list(), # historical y
                 'a': list(), # list of a known k steps in advance
                 't': list(),
                 'k': k}
        else:
            # Assert immutability of k, dimensions of y,a
            if s['y']:
                assert len(y) == len(s['y'][0])
                assert k == s['k']
            if s['a']:
                assert len(a) == len(s['a'][0])

        if y is None:
            return None, s, None
        else:
            s['y'].append(y)
            if a is not None:
                s['a'].append(a)
            if t is not None:
                assert isinstance(t,float), 'epoch time please'
                s['t'].append(t)

            if len(s['y']) > max(2 * k + 5, PROPHET_META['n_warm']):
                # Offset y, t, a are supplied to prophet interface
                t_arg = s['t'][k:] if t is not None else None
                a_arg = s['a']
                y_arg = s['y'][k:]
                x, x_std, forecast, model = nprophet_iskater_factory(y=y_arg, k=k, a=a_arg, t=t_arg,
                                                                    freq=freq, n_max=n_max,
                                                                    recursive=recursive, model_params=model_params)
                s['m'] = True # Flag indicating a model has been fit (there is no point keeping the model itself, however)
            else:
                x = [y[0]] * k
                x_std = None

            # Get running mean prediction errors from the prediction parade
            x_resid, x_resid_std, s['p'] = parade(p=s['p'], x=x, y=y[0])
            x_resid = nonecast(x_resid,y[0])
            x_resid_std = nonecast(x_resid_std,1.0)

            # Compute center of mass between bias-corrected and uncorrected predictions
            x_corrected = np.array(x_resid) + np.array(x)
            x_center = nonecenter(m=[emp_mass, 1 - emp_mass], x=[x_corrected, x])
            x_std_center = nonecenter(m=[emp_std_mass, 1 - emp_std_mass], x=[x_resid_std, x_std])

            return x_center, x_std_center, s


if __name__=='__main__':
    from timemachines.skatertools.data.real import hospital
    from timemachines.skatertools.evaluation.evaluators import evaluate_mean_absolute_error

    k = 5
    y = hospital(n=200)
    f = nprophet_skater_factory
    err2 = evaluate_mean_absolute_error(f=f, k=k, y=y, n_burn=50)
    print(err2)



from timemachines.skatertools.data.real import hospital_with_exog
from timemachines.skatertools.visualization.priorplot import prior_plot, prior_plot_exogenous
import matplotlib.pyplot as plt
import numpy as np


def hospital_prior_plot_exogenous(f, k=None, x0=np.nan, n=150, n_plot=25):
    y, a = hospital_with_exog(k=k, n=450, offset=True)
    err = prior_plot_exogenous(f=f,k=k,y=y,a=a, n=450, n_plot=50)







from timemachines.skaters.dlm.dlmunivariate import dlm_univariate_r3, dlm_univariate_a, dlm_univariate_b
from timemachines.skaters.dlm.dlmexogenous import dlm_exogenous_a, dlm_exogenous_b, dlm_exogenous_r3
from timemachines.skaters.dlm.dlmcomposed import DLM_SKATERS_COMPOSED

DLM_SKATERS = [ dlm_univariate_a, dlm_univariate_b ] + DLM_SKATERS_COMPOSED
DLM_R3_SKATERS = [ dlm_univariate_r3, dlm_exogenous_r3 ]


from timemachines.skatertools.utilities.conventions import positive_log_scale, to_space

# Some "global" prophet defaults set outside the model

PROPHET_META={'n_warm':400,  # Default minimum number of data points before fitting
              'freq':'H',    # Sampling frequency used if none is supplied, or t not supplied
              'n_max':500}   # Maximum length of training data

# Some prophet model parameters

PROPHET_MODEL = dict(changepoint_prior_scale=0.05,  # Suggested  [0.001, 0.01, 0.1, 0.5],
                    seasonality_prior_scale = 10.0,  # Suggested  [0.01, 0.1, 1.0, 10.0],
                    holidays_prior_scale = 10.0,  # Suggested  [0.01, 0.1, 1.0, 10.0]
                    seasonality_mode = 'additive',  # ['additive','multiplicative']
                    changepoint_range = 0.8,  # Suggested  [0.8-0.95]
                    interval_width = 0.34134*2)   # So it is +/- one standard deviation

PROPHET_MODEL_LOG_LOW = dict(changepoint_prior_scale=0.001,  # Suggested  [0.001, 0.01, 0.1, 0.5],
                    seasonality_prior_scale = 0.01,  # Suggested  [0.01, 0.1, 1.0, 10.0],
                    holidays_prior_scale = 0.01)  # Suggested  [0.01, 0.1, 1.0, 10.0]


PROPHET_MODEL_LOG_HIGH = dict(changepoint_prior_scale=0.5,  # Suggested  [0.001, 0.01, 0.1, 0.5],
                    seasonality_prior_scale = 20.,  # Suggested  [0.01, 0.1, 1.0, 10.0],
                    holidays_prior_scale = 20.,  # Suggested  [0.01, 0.1, 1.0, 10.0]
                    changepoint_range = 0.95)  # Suggested  [0.8-0.95]

PROPHET_MODEL_LINEAR_LOW  = dict(changepoint_range = 0.8)  # Suggested  [0.8-0.95]
PROPHET_MODEL_LINEAR_HIGH = dict(changepoint_range = 0.8)  # Suggested  [0.8-0.95]


def prophet_params(r:float,dim:int, param_names:[str])->dict:
    """ Interpret r in (0,1) as dict of param values """
    u = to_space(r, dim=dim)
    return dict([(name, prophet_param(name, ui)) for name, ui in zip(param_names, u)])


def prophet_param(param_name, u:float):
    """ Map u in (0,1) to sensible parameter value """
    assert 0 <= u <= 1
    if param_name in PROPHET_MODEL_LINEAR_LOW:
        lb = PROPHET_MODEL_LINEAR_LOW[param_name]
        ub = PROPHET_MODEL_LINEAR_HIGH[param_name]
        return (ub-lb) * u + lb
    elif param_name in PROPHET_MODEL_LOG_LOW:
        lb = PROPHET_MODEL_LOG_LOW[param_name]
        ub = PROPHET_MODEL_LOG_HIGH[param_name]
        return positive_log_scale(u=u,low=lb,high=ub)
    else:
        raise ValueError()



class ProphetWarning(UserWarning):
    pass


from typing import Callable
from timemachines.skatertools.utilities.conventions import Y_TYPE, K_TYPE, A_TYPE, T_TYPE, E_TYPE, R_TYPE, targets
import numpy as np


###########################################################################
#                                                                         #
#  Sequential assimilation of observations using a skater, strongly       #
#  suggesting what a skater function is supposed to accomplish, and how   #
#  it should behave                                                       #
#                                                                         #
#  See also skatertools.evaluation.evaluators                             #
#           skatertools.visualization.priorplot.prior_plot                #
#                                                                         #
###########################################################################

def posterior(f:Callable,       # "Model" to run forward through observations
              y:[Y_TYPE],       # Observations
              k:K_TYPE=1,       # Steps ahead to predict, typically integer
              a:[A_TYPE]=None,  # Data known in advance, or maybe action-conditional variables
              t:[T_TYPE]=None,  # Times of observations (epoch seconds)
              e:E_TYPE=None,    # Computation time limit per observation, and fit hint
              r:R_TYPE=None):   # Hype(r)-pa(r)amete(r)s squished into [0,1]

    """ Computes all k-step ahead estimates given history

         returns:  x:     [ [float] ]    list of all prediction vectors
                   x_std: [ [float] ]    list of all std vectors
    """
    s = dict()           # <- Skater will receive empty dictionary on the first invocation
    a = [None for _ in y] if a is None else a
    t = [None for _ in y] if t is None else t
    if r is not None:
        assert 0 <= r <= 1
    x = list()
    x_std = list()
    for yi, ai, ti in list(zip(y, a, t)):
        if r is None:
            xi, x_stdi, s = f(y=yi, s=s, k=k, a=ai, t=ti, e=e) # <-- Must provide k-vectors (mean, std devo)
        else:
            xi, x_stdi, s = f(y=yi, s=s, k=k, a=ai, t=ti, e=e, r=r)
        x.append(xi)
        x_std.append(x_stdi)
    return x, x_std


def prior(f, y:[Y_TYPE], k:int, a:[A_TYPE]=None, t:[T_TYPE]=None, e:E_TYPE=None, r:float=None,
          x0:float=np.nan, std0:float=np.nan):
    """ Compute k-step ahead estimates
        Align predictions with the observations y, for easy error computation
    """
    assert k is not None
    x_post, x_std = posterior(f=f, y=y, k=k, a=a, t=t, e=e, r=r)
    x_prior = [[x0]*k]*k + list(x_post)[:-k]
    x_prior_std = [[std0]*k]*k + list(x_post)[:-k]
    return x_prior, x_prior_std


def residuals(f, y:[Y_TYPE], k:int, a=None, t=None, e=None, r=0.5, x0=np.nan, n_burn=50):
    assert n_burn>k
    x, _ = prior(f=f, y=y, k=k, a=a, t=t, e=e, r=r, x0=x0)
    yt = targets(y)
    xk = [ xt[-1] for xt in x]
    return np.array(xk[-n_burn:])-np.array(yt[-n_burn:])



from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any
from timemachines.skatertools.components.residuals import residual



def residual_chaser_factory(y :Y_TYPE, s:dict, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None,
                            f1=None, f2=None, r1=None, r2=None)->([float] , Any , Any):
    """ Second model predicts k=1, k=k residuals of the first, and interpolates

          f1  - A skater making the primary prediction
          f2  - A skater designed to predict residuals ... both 1 step ahead and k-steps ahead
          r1  - hyper-params for f1, if any
          r2  - hyper-params for f2, if any

       It *may* make sense to choose an f2 that shrinks towards zero.
    """
    if k == 1:
        J = [1]
    else:
        J = [1,k]  # Determines horizons over which residual model is used.
                   # We'd rather not call the residual model k-times

    y0 = wrap(y)[0]
    if not s.get('s1'):
        s = {'sres': {},                      # Residual state ... used to determine the residual
             'x': y0,
             's1':{},                         # First model state
             's2':dict([(j,{}) for j in J]),  # Residual model states
             'n_obs':0}

    if y0 is None:
        return None, None, s
    else:
        # Use the first skater to predict
        if r1 is None:
            x1, x1_std, s['s1'] = f1(y=y,s=s['s1'],k=k, a=a,t=t,e=e)
        else:
            x1, x1_std, s['s1'] = f1(y=y, s=s['s1'], k=k, a=a, t=t, e=e, r=r1)
        resid1, s['sres'] = residual(s['sres'],y=y0,x=x1)

        s['n_obs']+=1

        # Use the second skater to predict j-step ahead residuals
        # There are two copies of the residual model employed.
        res_j_hat = [None for j in J]
        res_j_std = [None for j in J]
        for jpos,j in enumerate(J):
            j_ahead_residual = resid1[j-1]
            if r2 is None:
                _x, _std, s['s2'][j] = f2(y=j_ahead_residual, s=s['s2'][j], k=j, a=a, t=t, e=e)
            else:
                _x, _std, s['s2'][j] = f2(y=j_ahead_residual, s=s['s2'][j], k=j, a=a, t=t, e=e,r=r2)
            res_j_hat[jpos] = _x[jpos]
            res_j_std[jpos] = _std[jpos]

        # Interpolate
        if k==1:
            res_interp = res_j_hat
            res_interp_std = res_j_std
        else:
            import numpy as np
            ks = list(range(1,k+1))
            res_interp = np.interp( x=ks, xp=J, fp=res_j_hat )
            res_interp_std = np.interp(x=ks, xp=J, fp=res_j_std)

        # Residual   res =  y - x1,   so  x1+res ~ y  .... one hopes
        x_hat = [ resj+x1j for resj, x1j in zip( res_interp, x1) ]
        return x_hat, res_interp_std, s













import numpy as np
import matplotlib.pyplot as plt

# TODO

def mesh2d_(f,*args):
    """ Plot function taking len 2 vector as single argument
          f(xs)
    """
    def g(x,y,*args):
        return f(np.array([x,y]),*args)
    mesh2d(g,*args)


def mesh2d(f,*args):
    """ Plot function taking two arguments
        f(x,y)
    """
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    x = y =  np.arange(-0.5, 0.5, 0.005)
    X, Y = np.meshgrid(x, y)
    zs = np.array([ f(x_,y_,*args) for x_,y_ in zip( np.ravel(X), np.ravel(Y)) ])
    Z = zs.reshape(X.shape)

    ax.plot_surface(X, Y, Z)

    ax.set_xlabel('X Label')
    ax.set_ylabel('Y Label')
    ax.set_zlabel('Z Label')

    plt.show()


import numpy as np
import divinity as dv
from timemachines.skatertools.utilities.conventions import Y_TYPE, K_TYPE
from timemachines.skatertools.utilities.suppression import no_stdout_stderr
from timemachines.skatertools.visualization.priorplot import prior_plot
from timemachines.skatertools.utilities.conventions import wrap
from timemachines.skaters.divine.divineparams import MIN_N_WARM, DIVINE_MODEL
from copy import deepcopy

# This guy is a little chatty


def divinity_univariate_factory(y:Y_TYPE, s, k:K_TYPE, a=None, t=None, e=None,
                                max_buffer_len=1000,
                                n_warm = 101,
                                model_params:dict=None):
    """ A partial wrapping of the divinity library with notable limitations:

         - Fits every invocation
         - Ignores exogenous variables
         - State is merely a buffer

    """
    y0 = wrap(y)[0]
    assert n_warm>=101,' You must use n_warm'

    if not s:
        s = dict(y=[])

    if y0 is None:
        return None, None, s   # Ignore suggestion to fit offline

    # Update buffer
    s['y'].append(y0)
    if len(s['y'])>max_buffer_len+1000:
        s['y'] = s['y'][-max_buffer_len:]

    # Fit and predict, if warm, or just last value
    if len(s['y']) < max( n_warm, MIN_N_WARM ):
        return [y0]*k, [abs(y0)]*k, s
    else:
        with no_stdout_stderr():
            kwargs = deepcopy(DIVINE_MODEL)
            if model_params:
                kwargs.update(**model_params)
            model = dv.divinity(forecast_length=k,**kwargs)
            model.fit(np.array(s['y']))
        x = list(model.predict())
        x_std = [1.0]*k # TODO: fixme
        return x, x_std, s


if __name__=='__main__':
    err = prior_plot(f=divinity_univariate_factory, k=5, n=500)
    pass




from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any

# Similar to basic skaters, but no maintenence of empirical errors is performed
# Useful as components when you just need a point estimate for some secondary quantity


def trivial_last_value(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None )->([float] , [float] , Any):
    """ Last value cache """

    if y is None:
        return None, None, s
    else:
        y0 = wrap(y)[0]       # Ignore the rest
        x = [y0]*k            # What a great prediction !
        return x, 1.0, {}


def trivial_ema_r1(y :Y_TYPE, s, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None, r:R_TYPE=None):
    """ Exponential moving average
          r      weight to place on existing anchor point
    """
    assert r is not None
    y0 = wrap(y)[0]
    if not s.get('rho'):
        s = {'x':y0,
             'rho':r}
        assert 0 <= s['rho'] <= 1, 'Expecting rho=r to be between 0 and 1'
    else:
        assert abs(r-s['rho'])<1e-6,'rho=r is immutable'

    if y0 is None:
        return None, s, None
    else:
        s['x'] = s['rho']*s['x'] + (1-s['rho'])*y0         # Make me better !
        x = [ s['x']*k ]
        return x, [1.0]*k, s


if __name__=='__main__':
    from timemachines.skatertools.data import hospital_with_exog
    from timemachines.skatertools.evaluation.evaluators import evaluate_mean_absolute_error

    k = 3
    y, a = hospital_with_exog(k=k,n=500)
    f = trivial_ema_r1
    err1 = evaluate_mean_absolute_error(f=f, k=k, y=y, a=a, r=0.9, n_burn=50)




from timemachines.skaters.simple.movingaverage import EMA_SKATERS, EMA_R1_SKATERS
from timemachines.skaters.simple.linear import LINEAR_SKATERS
from timemachines.skaters.divine.divineskaters import DIVINE_SKATERS
from timemachines.skaters.proph.allprophetskaters import PROPHET_SKATERS, PROPHET_R2_SKATERS
from timemachines.skaters.dlm.alldlmskaters import DLM_SKATERS
from timemachines.skaters.simple.thinking import THINKING_SKATERS
from timemachines.skaters.pmd.allpmdskaters import PMD_SKATERS
from timemachines.skaters.tsa.alltsaskaters import TSA_SKATERS
from timemachines.skaters.nproph.allnprophetskaters import NPROPHET_UNIVARIATE_SKATERS

SKATERS_R3 = []
SKATERS_R2 = PROPHET_R2_SKATERS
SKATERS_R1 = EMA_R1_SKATERS


# And with no hyper-params...
SKATERS = EMA_SKATERS + LINEAR_SKATERS + PROPHET_SKATERS + DIVINE_SKATERS + DLM_SKATERS + \
          THINKING_SKATERS + PMD_SKATERS + TSA_SKATERS + NPROPHET_UNIVARIATE_SKATERS


def skater_from_name(name):
    valid = [f for f in SKATERS if f.__name__==name ]
    return valid[0] if len(valid)==1 else None


from timemachines.skaters.pmd.pmdskaterfactory import pmd_skater_factory
from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE,T_TYPE,E_TYPE, wrap
from timemachines.skaters.simple.movingaverage import empirical_last_value
from timemachines.skatertools.visualization.priorplot import prior_plot, prior_plot_exogenous


def pmd_exogenous(y:Y_TYPE, s:dict, k:int=1, a:A_TYPE=None, t:T_TYPE=None, e:E_TYPE=None):
    """ Predict using auto_arima, with both simultaneously observed and known in advance variables
        This skater has no hyper-parameters

        y: Y_TYPE    scalar or list where y[1:] are interpreted as contemporaneously observed exogenous variables
        s:           state
        k:           Number of steps ahead to predict
        a:           (optional) scalar or list of variables known k-steps in advance.
                      (IMPORTANT: If supplying 'a', provide the known variable k steps ahead, not the contemporaneous one !).
        t:           (optional) Time of observation.
        e:           (optional) Maximum computation time (supply e>60 to give hint to do fitting)

        :returns: x [float] , s', scale [float]
    """
    return pmd_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, method='auto')


def pmd_known(y:Y_TYPE, s:dict, k:int=1, a:A_TYPE=None, t:T_TYPE=None, e:E_TYPE=None):
    """ Uses known-in-advance but not y[1:] """
    y0 = [wrap(y)[0]]
    return pmd_skater_factory(y=y0, s=s, k=k, a=a, t=t, e=e, method='auto')


def pmd_univariate(y:Y_TYPE, s:dict, k:int=1, a:A_TYPE=None, t:T_TYPE=None, e:E_TYPE=None):
    """ Uses only y[0] and ignores y[1:] and a[:] """
    y0 = [wrap(y)[0]]
    return pmd_skater_factory(y=y0, s=s, k=k, a=None, t=t, e=e, method='auto')


def pmd_exog_compare(f,k=1):
    from timemachines.skatertools.evaluation.evaluators import evaluate_mean_absolute_error
    from timemachines.skatertools.evaluation.evaluators import hospital_with_exog
    y, a = hospital_with_exog(k=k)
    y0 = [ yi[0] for yi in y ]

    r = 0.1 # Doesn't matter?
    err1 = evaluate_mean_absolute_error(f=f, k=k, y=y0, r=r, n_burn=250)
    err2 = evaluate_mean_absolute_error(f=f, k=k, y=y, r=r, n_burn=250)
    err3 = evaluate_mean_absolute_error(f=f, k=k, y=y, r=r, a=a, n_burn=250)
    errlv = evaluate_mean_absolute_error(f=empirical_last_value, k=k, y=y, r=r, a=a, n_burn=250)


    print('----------------')
    print("Error w/o exogenous   = "+str(err1))
    print("Error w   exogenous   = "+str(err2))
    print("Error w   exo + known = "+str(err3))
    print("Error last val cache  = " + str(errlv))


if __name__ == '__main__':
    f = pmd_exogenous
    if True:
        prior_plot_exogenous(f=f, k=1, n=200)
    if True:
        prior_plot(f=f,k=1,n=200)



from pydlm import dlm, trend, seasonality
from timemachines.skatertools.utilities.conventions import dimension, to_int_log_space, split_exogenous, Y_TYPE, R_TYPE
from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous
import math
import numpy as np
import matplotlib.pyplot as plt
from timemachines.skaters.dlm.dlmhacks import fixedAutoReg


###################################################
#                                                 #
#      Univariate skater based on pydlm           #
#                                                 #
###################################################


def dlm_univariate_a(y, s: dict, k: int, a=None, t=None, e=None):
    ra = 0.25
    return dlm_univariate_r3(y=y, s=s, k=k, r=ra)


def dlm_univariate_b(y, s: dict, k: int, a=None, t=None, e=None):
    ra = 0.75
    return dlm_univariate_r3(y=y, s=s, k=k, r=ra)


def dlm_univariate_r3(y, s: dict, k: int, a=None, t=None, e=None, r=None):
    """ Univariate filter

            - Uses the discounting method of H/W so, doesn't need to be fit as often
            - Discount factors are periodically tuned
            - The hyper-parameter controls 'auto_degree', 'trend_degree',  'period'

        :returns: x, x_std, s
    """
    assert r is not None, 'Requires hyper-parameter (interpreted in dimension 3) '
    if not s:
        s = dict()
        s = dlm_set_univariate_params(s=s, r=r)
        s['dim'] = dimension(y)
        s['n_obs'] = 0
        s['model'] = dlm([], printInfo=False) + trend(s['trend_degree'], s['discount']) + seasonality(s['period'],
                                                                                                      s['discount'])
        s['model'] = s['model'] + fixedAutoReg(degree=s['auto_degree'], name='ar', w=1.0)

    if y is not None:
        s['n_obs'] += 1
        assert isinstance(y, float) or len(y) == s['dim'], ' Cannot change dimension of input in flight '
        y0, exog = split_exogenous(y=y)
        y0_passed_in = None if np.isnan(y0) else y0  # pydlm uses None for missing values
        s['model'].append([y0_passed_in])
        num_obs = len(s['model'].data) if s.get('model') else 0
        if num_obs % s['n_fit'] == s['n_fit'] - 1:
            # Perform periodic tuning of discount factors
            _, _, s = dlm_univariate_r3(y=None, s=s, k=k, a=a, t=t, e=1000, r=r)
        s['model'].fitForwardFilter()
        return _dlm_prediction_helper(s=s, k=k, y=y)

    if y is None and e > 60:
        s['model'].tune()  # Tunes discount factors
        s['model'].fit()
        return None, None, s


def _dlm_prediction_helper(s, k:int, y:Y_TYPE):
    """ Calls down to predictN, if we are passed the warm-up stage """
    num_obs = len(s['model'].data) if s.get('model') else 0
    if num_obs < s['n_warm']:
        y0, _ = split_exogenous(y)
        return [y0]*k, [abs(y0)]*k, s
    else:
        x_mean, x_var = s['model'].predictN(N=k)
        x = list(x_mean)
        x_std = [ math.sqrt(v) for v in x_var]
        return x, x_std, s



def dlm_set_univariate_params(s, r:R_TYPE):
    period_choices = [3,5,7,10,12,16,24,32]
    s['auto_degree'], s['trend_degree'], period_choice = to_int_log_space(r,bounds=[
        (0.5,3.2),    # auto-reg degree. 1 means linear trend, 2 quadratic
        (0.1,3.4),    # seasonality degree
        (0,len(period_choices)) # seasonality period
    ])
    s['discount'] = 1 - ( r % 0.05 ) # Shouldn't matter too much as these get tuned eventually
    s['period'] = period_choices[period_choice]
    s['n_warm'] = 100
    s['n_fit']  = 500  # How often to tune discounts subsequently
    return s


if __name__ == '__main__':
    err = prior_plot_exogenous(f=dlm_univariate_r3, k=2, n=500, r=np.random.rand())
    plt.figure()
    print('done')
    pass



from pydlm import trend, seasonality, dynamic
from timemachines.skatertools.utilities.conventions import dimension, to_int_log_space, split_exogenous, Y_TYPE, R_TYPE
from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous
from timemachines.skaters.dlm.dlmhacks import quietDlm, fixedAutoReg
from timemachines.skatertools.utilities.conventions import wrap
import math
import numpy as np
import matplotlib.pyplot as plt


def dlm_exogenous_a(y, s: dict, k: int, a=None, t=None, e=None):
    ra = 0.25
    return dlm_exogenous_r3(y=y, s=s, k=k,  a=a, t=t, e=e, r=ra)


def dlm_exogenous_b(y, s: dict, k: int, a=None, t=None, e=None):
    rb = 0.75
    return dlm_exogenous_r3(y=y, s=s, k=k, a=a, t=t, e=e, r=rb)


def dlm_exogenous_r3(y, s, k, a, t, e, r):
    """ One way to use dlm
        :returns: x, s', w
    """
    if not s:
        s = dict()
        s['dim'] = dimension(y)
        s = dlm_set_exog_hyperparams(s=s, r=r)
        y0, exog = split_exogenous(y=y)
        s['n_obs'] = 0
        s['model'] = quietDlm([], printInfo=False) + trend(s['trend_degree'], s['discount']) + seasonality(s['period'], s['discount'])
        s['model'] = s['model'] + fixedAutoReg(degree=s['auto_degree'], name='ar', w=1.0)
        if exog:
            exog_wrapped = [[None if np.isnan(ex0) else ex0 for ex0 in exog]]
            s['model'] = s['model'] + dynamic(features=exog_wrapped, discount=0.99, name='exog') # Set's first exog

    if y is not None:
        y = wrap(y)
        assert dimension(y)==s['dim'],'Cannot change dimension of data sent'
        s['n_obs'] += 1
        y0, exog = split_exogenous(y=y)
        y0_passed_in = None if np.isnan(y0) else y0  # pydlm uses None for missing values
        s['model'].append([y0_passed_in])
        if exog:
            exog_wrapped = [[ None if np.isnan(ex0) else ex0 for ex0 in exog ]]
            if s['n_obs']>1:
                s['model'].append(data=exog_wrapped, component='exog') # Don't get first exog twice
        num_obs = len(s['model'].data) if s.get('model') else 0
        if num_obs % s['n_fit'] == s['n_fit']-1:
            _, _, s = dlm_exogenous_r3(y=None, s=s, k=k, a=a, t=t, e=10, r=r)
        s['model'].fitForwardFilter()
        return _dlm_exog_prediction_helper(s=s, k=k, y=y)

    if y is None:
        if dimension(y)==1:
            s['model'].tune(maxit=20)
            # Don't tune if exogenous ... haven't got this to work
        s['model'].fit()
        return None, None, s



def _dlm_exog_prediction_helper(s, k:int, y:Y_TYPE):
    num_obs = len(s['model'].data) if s.get('model') else 0
    if num_obs < s['n_burn']:
        y0, exog = split_exogenous(y)
        return [y0]*k, [abs(y0)]*k, s
    else:
        assert k==1,'only k==1 for now' # TODO: Fix to allow for k>1
        y0, exog = split_exogenous(y)
        if exog:
            exog_passed_in = [None if np.isnan(ex0) else ex0 for ex0 in exog]
            x_mean, x_var = s['model'].predict(featureDict={'exog':exog_passed_in[0]})
        else:
            x_mean, x_var = s['model'].predict()
        x = [x_mean[0, 0]]
        x_std = [ math.sqrt( x_var[0,0] ) ]

        return x, x_std, s


def dlm_set_exog_hyperparams(s, r:R_TYPE):
    # Univariate model with autoregressive components
    # This uses the discounting method of H/W so doesn't need to be fit as often
    period_choices = [3,5,7,10,12,16,24,32]
    s['auto_degree'], s['trend_degree'], period_choice = to_int_log_space(r,bounds=[
        (0.5,3.2),    # auto-reg degree. 1 means linear trend, 2 quadratic
        (0.1,3.4),    # seasonality degree
        (0,len(period_choices)) # seasonality period
    ])
    s['discount'] = 1 - ( r % 0.05 ) # Shouldn't matter too much as these get tuned eventually
    s['period'] = period_choices[period_choice]
    s['n_burn'] = 100
    s['n_fit']  = 500  # How often to tune discounts subsequently
    return s



if __name__ == '__main__':
    err = prior_plot_exogenous(f=dlm_exogenous_r3, k=1, n=1000, r=np.random.rand())
    plt.figure()
    print('done')
    pass



from typing import Union
from numpy import asarray
from itertools import zip_longest


def nonelen(l):
    return 0 if l is None else len(l)


def noneratio(x,y):
    return None if x is None or y is None else x/y


def nonecast(x,fill_value=0.0):
    if x is not None:
        return [xj if xj is not None else fill_value for xj in x]


def notallnone(x):
    return not all([ xj is None for xj in x])


def nearlysame(x,y,tol=1e-6):
    return all( ( abs(xj-yj)<tol for xj,yj in zip_longest(x,y,fillvalue=.777)))


def nonennearlysame(x,y,tol=1e-6):
    x_cast = nonecast(x, fill_value=333.0)
    y_cast = nonecast(x, fill_value=333.0)
    return nearlysame(x_cast,y_cast)


def nonecenter(m,x):
    """
    :param m:  List of masses
    :param x:  List of points, can be None or vector
    :returns: point that is center of mass
    """
    assert notallnone(m)
    assert notallnone(x)
    m_when_x_not_none = [ m for m,x in zip(m,x) if x is not None]
    x_when_x_not_none = [ x for m,x in zip(m,x) if x is not None]
    m_cast = nonecast(m_when_x_not_none,fill_value=0.0)
    x_cast = [ nonecast(pj,fill_value=0.0) for pj in x_when_x_not_none]
    if sum(m_cast)<1e-9:
        m_cast = [1./len(m_cast) for _ in m_cast ]
    return center(m_cast, x_cast)


def center(m:[float], x:[[float]])->[Union[None, float]]:
    """ Center of mass
          m   masses
          x   set of points
    """
    if len(m)!=len(x):
        import warnings
        warnings.warn('length mismatch')
    P = asarray([ asarray(xj) for xj in x])
    M = asarray(m)
    PM = (P.T * M).T
    try:
        PM_sum = PM.sum(axis=0)
    except:
        k = 1
        pass
    c = PM_sum / M.sum()
    return c.tolist()




from timemachines.skaters.pmd.pmdskaters import pmd_known, pmd_exogenous, pmd_univariate
from timemachines.skaters.pmd.pmdcomposed import pmd_exogenous_hypocratic, pmd_univariate_hypocratic

PMD_SKATERS = [pmd_univariate, pmd_exogenous_hypocratic]  # for now


import math


def weighted_average(y,w):
    return sum(wj * yj for yj, wj in zip(y, w)) / sum(w)


def normalize(w):
    return [ wj/sum(w) for wj in w]


def precision_weighted_skater(y,s,k,a,t,e,r=0.5):
    """

       Treats y[1],y[3],y[5]... as unbiased estimates of y[0]
       Treats y[2],y[4],...     as std

       r -   determines the exponent:

             r=0.0 corresponds to equal weighting independent of x_std supplied by the models
             r=0.5 corresponds to simple weighting and is the default
             r->1  will use only the most accurate skater in the ensemble

    """
    tol = 1e-6  # minimum allowed x_std
    expon = 2*math.atanh(r)/math.atanh(0.5) if r<1-1e-6 else 10.0
    J  = int((len(y)-1)/2)
    y_stds = [ tol + y[2*j+2] for j in range(J) ]
    w  = normalize( [ 1./math.pow(y_std,expon) for y_std in y_stds ] )
    y_hats = [ y[2*j+1] for j in range(J) ]
    x = weighted_average(y=y_hats, w=w)
    x_std = min(y_stds)  # Conservative - they probably are not independent
    x_interp = [y[0]+(j+1)/k*(x-y[0]) for j in range(k)]
    x_std_interp = [math.sqrt(j+1)*x_std for j in range(k)]
    return x_interp, x_std_interp, {}



from timemachines.skaters.pmd.pmdskaters import pmd_exogenous
from timemachines.skatertools.data import hospital_with_exog
from pmdarima.arima import ARIMA


# So the task here is to serialize and re-inflate the arima model, and thus all the skater state, to JSON
# It may be a lost cause, but by all means prove me wrong.


def example_pmd():
    s = {}
    y, a = hospital_with_exog(k=3)
    x = [ pmd_exogenous(y=yj,s=s,k=3,a=aj) for yj,aj in zip(y[:500],a) ]
    return s


def arima_res_to_dict(arima_res):
    state = arima_res.__dict__
    return state


def pmd_to_dict(pmd):
    pmd['model'] = pmd['model'].__getstate__()
    pmd['model']['arima_res_'] = arima_res_to_dict(pmd['model']['arima_res_'])
    return pmd


def pmd_from_dict(pmd):
    pmd['model']['arima_res_'] = ''


if __name__=='__main__':
    pmd = example_pmd()
    model = pmd['model']
    model1 = ARIMA(**model.get_params())
    prms = model.__dict__['arima_res_'].__dict__['_results'].params







import pmdarima as pm
from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE, S_TYPE
from timemachines.skatertools.components.observance import observance
from timemachines.skatertools.utilities.conventions import split_exogenous, wrap, dimension
from timemachines.skaters.pmd.pmddefaultparams import pmd_params
import numpy as np
from typing import Any
from timemachines.skatertools.components.chronometer import tick, tock, tocks
from pprint import pprint
from typing import Union, List


###################################################################################################
#                                                                                                 #
#                      PMD-ARIMA skater factory                                                   #
#                                                                                                 #
# As with all skaters, the intent is that you can cycle through observations as follows:          #
#                                                                                                 #
#      for yi, ai in zip(y, a):                                                                   #
#         x, x_std, s = pmd_exogenous(y=yi, s=s, k=7, a=ai)                                       #
#                                                                                                 #
# where each x and s_std are length 7, in this example, and a[0] is contemporaneous with y[7]     #
#                                                                                                 #
# Advantages:                                                                                     #
#       - Pretty good amortized invocation time, say if you fit every 100 invocations             #
# Disadvantages:                                                                                  #
#       - Model state is opaque or at best obscure                                                #
#                                                                                                 #
###################################################################################################


def pmd_skater_factory(y:Y_TYPE, s:dict, k:int=1, a:A_TYPE=None, t:T_TYPE=None, e:E_TYPE=None,
                       method: str= 'default', n_warm=50,
                       model_params:dict=None)->(Union[List[float],None],
                                                 Union[List[float],None], Any):
    """ Predict using both simultaneously observed and known in advance variables
        y: Y_TYPE    scalar or list where y[1:] are interpreted as contemporaneously observed exogenous variables
        s:           state
        k:           Number of steps ahead to predict
        a:           (optional) scalar or list of variables known k-steps in advance.
                     When calling, provide the known variable k steps ahead, not the contemporaneous one.
        t:           (optional) Time of observation.
        e:           (optional) Maximum computation time (supply e>60 to give hint to do fitting)

        :returns: x [float] , s', scale [float]

        Remarks:
           - Model params cannot be changed after the first invocation.
           - Allows y=None to be used
    """
    y = wrap(y)
    a = wrap(a)

    if not s.get('n_obs'):
        # Initialize
        s['n_obs'] = 0
        s['model'] = None
        s['immutable'] = pmd_set_immutable(k=k, y=y, a=a, n_warm=n_warm)
        s['params'] = pmd_params(method=method)
        if model_params:
            s['params'].update(model_params)
        s['o'] = dict()                         # Observance
    else:
        pmd_check_consistent_usage(y=y,s=s,a=a,k=k)

    tick(s)
    if t is not None:
        pass # Other models might perform an evolution step here. Not applicable to PMDARIMA

    if y is not None:
        # Receive observation y[0], possibly exogenous y[1:] and possibly k-in-advance a[:]
        # Collect from queues the contemporaneous variables
        s['n_obs']+=1
        y_t, z = split_exogenous(y)
        x_t, s['o'] = observance(y=y,o=s['o'],k=k,a=a)

        # Update the pmdarima model itself
        if x_t is not None:
            if s['model'] is not None:
                if x_t:
                    s['model'].update([y_t], [x_t])
                else:
                    s['model'].update([y_t])

        # Predict
        if s['model'] is None:
            # Fall back to last value if there is no model calibrated as yet
            x = [y_t]*k
            if len(s['o']['x']) > 5 + 2*k:
                Y = s['o']['y'][k+1:]
                X = s['o']['x'][k+1:]
                x_std = [ np.nanstd( [ xi[0]-yk[0] for xi, yk in zip( X, Y[j:] ) ] ) for j in range(1,k+1) ]
            else:
                x_std = [1.0]*k   # Fallback to dreadful estimate
        else:
            # Predict forward, supplying known data if it exists
            if not a and not z:
                z_forward = None
            else:
                if not a:
                    z_forward = [z]*k
                else:
                    z_forward = [ list(z) + list(ai) for ai in s['o']['a'] ]  # Add known k-steps ahead
                                    # This estimate could be improved by predicting z's and attenuating
                                    # It is only really a good idea for k=1
            x, ntvls = s['model'].predict(n_periods=k, X=z_forward, return_conf_int=True, alpha=s['immutable']['alpha'])
            x_std = list([ ntvl[1] - ntvl[0] for ntvl in ntvls ])

    # Fit
    tock(s)
    if pmd_it_is_time_to_fit(s=s, e=e):
        tick(s)
        X = s['o'].get('x') or None
        Y = s['o']['y']
        s['model'] = pm.auto_arima(y=Y, X=X, **s['params'])
        print(s['model'])
        print(s['model'].arima_res_.params)
        pprint(tocks(s))
        tock(s,'fit')
        pprint(tocks(s))

    if y is not None:
        return list(x), list(x_std), s
    else:
        return None, None, s


def pmd_it_is_time_to_fit(s:S_TYPE, e:E_TYPE)->bool:
    """ Provided 60 seconds, or getting stale """
    return s['n_obs'] == s['immutable']['n_warm'] or \
           (s['n_obs'] > s['immutable']['n_warm'] and ((e is not None and e>60) or s['n_obs'] % s['immutable']['n_fit'] == 0))


def pmd_set_immutable(y: Y_TYPE, k:int, a:A_TYPE=None, n_warm:int=20):
    """ Set on the first invocation, when s={} is passed """
    return {'k': k,
            'alpha':0.25,   # Determines confidence interval
            'n_fit':250,
            'n_warm':n_warm,
            'dim_exog': dimension(y) - 1,
            'dim_a': dimension(a)}


def pmd_check_consistent_usage(y:Y_TYPE,s,k,a):
    if y is not None:
        assert dimension(y)-1 == s['immutable']['dim_exog']
    if k is not None:
        assert k==s['immutable']['k']
    if a is not None:
        assert dimension(a)==s['immutable']['dim_a']







from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any
from timemachines.skatertools.components.parade import parade
from timemachines.skatertools.utilities.nonemath import nonecast
from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory
import math

# Fast elementary time series skaters, useful as benchmarks, pre-processing, sub-components etc
# They maintain running mean/std of their own empirical errors


def empirical_last_value(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """ Last value cache, with empirical std """

    if not s.get('p'):
        s = {'p':{}}   # Initialize prediction parade

    if y is None:
        return None, None, s
    else:
        y0 = wrap(y)[0]       # Ignore the rest
        x = [y0]*k            # What a great prediction !
        bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y0)  # update residual queue
        return x, x_std, s


def empirical_ema_r1(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None, r:R_TYPE=None):
    """ Exponential moving average, with empirical std

          r      weight to place on existing anchor point

    """
    assert r is not None
    y0 = wrap(y)[0]
    if not s.get('p'):
        s = {'p':{},
             'x':y0,
             'rho':r}
        assert 0 <= s['rho'] <= 1, 'Expecting rho=r to be between 0 and 1'
    else:
        assert abs(r-s['rho'])<1e-6,'rho=r is immutable'

    if y0 is None:
        return None, s, None
    else:
        s['x'] = s['rho']*s['x'] + (1-s['rho'])*y0         # Make me better !
        x = [s['x']]*k
        _we_ignore_bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y0)
        x_std_fallback = nonecast(x_std,fill_value=1.0)
        return [s['x']] * k, x_std_fallback, s


def sluggish_moving_average(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None):
    return empirical_ema_r1(y=y, s=s, k=k, a=a, t=t, e=e, r=0.99)


def slowly_moving_average(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None):
    return empirical_ema_r1(y=y, s=s, k=k, a=a, t=t, e=e, r=0.95)


def quickly_moving_average(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None):
    return empirical_ema_r1(y=y, s=s, k=k, a=a, t=t, e=e, r=0.75)


def rapidly_moving_average(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None):
    return empirical_ema_r1(y=y, s=s, k=k, a=a, t=t, e=e, r=0.5)

EMA_BASIC_SKATERS = [empirical_last_value, sluggish_moving_average, slowly_moving_average, quickly_moving_average, rapidly_moving_average ]


def precision_ema_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
         Precision weight several different moving averages
    """
    fs = EMA_BASIC_SKATERS
    return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.5)


def balanced_ema_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
         More evenly weighted
    """
    fs = EMA_BASIC_SKATERS
    return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.25)


def aggressive_ema_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
         Heavily weight the best
    """
    fs = EMA_BASIC_SKATERS
    return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.9)


EMA_ENSEMBLE_SKATERS = [ balanced_ema_ensemble, aggressive_ema_ensemble, precision_ema_ensemble]

EMA_SKATERS = EMA_BASIC_SKATERS + EMA_ENSEMBLE_SKATERS

EMA_R1_SKATERS = [empirical_ema_r1]


if __name__=='__main__':
    from timemachines.skatertools.data.real import hospital_with_exog
    from timemachines.skatertools.evaluation.evaluators import evaluate_mean_absolute_error

    k = 3
    y, a = hospital_with_exog(k=k,n=500)
    f = aggressive_ema_ensemble
    err1 = evaluate_mean_absolute_error(f=f, k=k, y=y, a=a, n_burn=50)













